/*
**    TP CPE Lyon
**    Copyright (C) 2013 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "matrix4.hpp"

#include "matrix3.hpp"
#include "vec3.hpp"
#include "vec4.hpp"
#include "../libcommon/exception_cpe.hpp"
#include <cmath>
#include <iostream>
#include <sstream>


namespace cpe
{

matrix4::matrix4()
{
    m[0]=1;m[4]=0;m[8]=0;m[12]=0;
    m[1]=0;m[5]=1;m[9]=0;m[13]=0;
    m[2]=0;m[6]=0;m[10]=1;m[14]=0;
    m[3]=0;m[7]=0;m[11]=0;m[15]=1;
}

matrix4::matrix4(float x00,float x01,float x02,float x03,
                 float x10,float x11,float x12,float x13,
                 float x20,float x21,float x22,float x23,
                 float x30,float x31,float x32,float x33)
{
    m[0]=x00;m[4]=x01;m[8]=x02;m[12]=x03;
    m[1]=x10;m[5]=x11;m[9]=x12;m[13]=x13;
    m[2]=x20;m[6]=x21;m[10]=x22;m[14]=x23;
    m[3]=x30;m[7]=x31;m[11]=x32;m[15]=x33;
}

matrix4::matrix4(const matrix3& m3)
{
    m[0]=m3(0,0);m[4]=m3(0,1);m[8]=m3(0,2);m[12]=0;
    m[1]=m3(1,0);m[5]=m3(1,1);m[9]=m3(1,2);m[13]=0;
    m[2]=m3(2,0);m[6]=m3(2,1);m[10]=m3(2,2);m[14]=0;
    m[3]=0;m[7]=0;m[11]=0;m[15]=1;
}

matrix4 matrix4::identity()
{
    return matrix4(1,0,0,0,
                   0,1,0,0,
                   0,0,1,0,
                   0,0,0,1);
}
matrix4 matrix4::zeros()
{
    return matrix4(0,0,0,0,
                   0,0,0,0,
                   0,0,0,0,
                   0,0,0,0);
}


matrix4 matrix4::rotation(const vec3& axis,float angle)
{
    vec3 n=axis.normalized();

    float cos_t=std::cos(angle);
    float sin_t=std::sin(angle);

    return matrix4(cos_t+n[0]*n[0]*(1-cos_t),n[0]*n[1]*(1-cos_t)-n[2]*sin_t,n[1]*sin_t+n[0]*n[2]*(1-cos_t),0,
                   n[2]*sin_t+n[0]*n[1]*(1-cos_t),cos_t+n[1]*n[1]*(1.0-cos_t),-n[0]*sin_t+n[1]*n[2]*(1-cos_t),0,
                   -n[1]*sin_t+n[0]*n[2]*(1-cos_t),n[0]*sin_t+n[1]*n[2]*(1-cos_t),cos_t+n[2]*n[2]*(1-cos_t),0,
                   0,0,0,1);

}

matrix4 matrix4::scale(float s)
{
    return matrix4(s,0,0,0,
                   0,s,0,0,
                   0,0,s,0,
                   0,0,0,1);
}

matrix4 matrix4::scale(float s_x,float s_y,float s_z,float s_w)
{
    return matrix4(s_x,0.0,0.0,0.0,
                   0.0,s_y,0.0,0.0,
                   0.0,0.0,s_z,0.0,
                   0.0,0.0,0.0,s_w);
}
matrix4 matrix4::translation(const vec3& translation)
{
    return matrix4(1,0,0,translation.x(),
                   0,1,0,translation.y(),
                   0,0,1,translation.z(),
                   0,0,0,1);
}

matrix4 matrix4::transformation(const matrix3& m3,const vec3& translation)
{
    return matrix4(m3(0,0),m3(0,1),m3(0,2),translation.x(),
                   m3(1,0),m3(1,1),m3(1,2),translation.y(),
                   m3(2,0),m3(2,1),m3(2,2),translation.z(),
                   0,0,0,1);
}

float matrix4::operator()(const size_t& k1,const size_t& k2) const
{
    assert_size(k1,k2);
    return m[4*k2+k1];
}
float& matrix4::operator()(const size_t& k1,const size_t& k2)
{
    assert_size(k1,k2);
    return m[4*k2+k1];
}

matrix4 matrix4::operator+(const matrix4& m2) const
{
    const float* pm2=m2.pointer();

    return matrix4(m[0]+pm2[0],m[4]+pm2[4],m[8]+pm2[8],m[12]+pm2[12],
                   m[1]+pm2[1],m[5]+pm2[5],m[9]+pm2[9],m[13]+pm2[13],
                   m[2]+pm2[2],m[6]+pm2[6],m[10]+pm2[10],m[14]+pm2[14],
                   m[3]+pm2[3],m[7]+pm2[7],m[11]+pm2[11],m[15]+pm2[15]);
}

matrix4 operator+(float s,const matrix4& m)
{
    const float* pm=m.pointer();
    return matrix4(pm[0]+s,pm[4]+s,pm[8]+s,pm[12]+s,
                   pm[1]+s,pm[5]+s,pm[9]+s,pm[13]+s,
                   pm[2]+s,pm[6]+s,pm[10]+s,pm[14]+s,
                   pm[3]+s,pm[7]+s,pm[11]+s,pm[15]+s);
}

matrix4 matrix4::operator+(float s) const
{
    return s+(*this);
}

matrix4 matrix4::operator-(const matrix4& m2) const
{
    const float* pm2=m2.m;

    return matrix4(m[0]-pm2[0],m[4]-pm2[4],m[8]-pm2[8],m[12]-pm2[12],
                   m[1]-pm2[1],m[5]-pm2[5],m[9]-pm2[9],m[13]-pm2[13],
                   m[2]-pm2[2],m[6]-pm2[6],m[10]-pm2[10],m[14]-pm2[14],
                   m[3]-pm2[3],m[7]-pm2[7],m[11]-pm2[11],m[15]-pm2[15]);
}

matrix4 operator-(float s,const matrix4& m)
{
    const float* pm=m.pointer();
    return matrix4(s-pm[0],s-pm[4],s-pm[8],s-pm[12],
                   s-pm[1],s-pm[5],s-pm[9],s-pm[13],
                   s-pm[2],s-pm[6],s-pm[10],s-pm[14],
                   s-pm[3],s-pm[7],s-pm[11],s-pm[15]);
}

matrix4 matrix4::operator-(float s) const
{
    return matrix4(m[0]-s,m[4]-s,m[8]-s,m[12]-s,
                   m[1]-s,m[5]-s,m[9]-s,m[13]-s,
                   m[2]-s,m[6]-s,m[10]-s,m[14]-s,
                   m[3]-s,m[7]-s,m[11]-s,m[15]-s);
}

matrix4 operator*(float s,const matrix4& m)
{
    const float* pm=m.pointer();
    return matrix4(pm[0]*s,pm[4]*s,pm[8]*s,pm[12]*s,
                   pm[1]*s,pm[5]*s,pm[9]*s,pm[13]*s,
                   pm[2]*s,pm[6]*s,pm[10]*s,pm[14]*s,
                   pm[3]*s,pm[7]*s,pm[11]*s,pm[15]*s);
}

matrix4 matrix4::operator*(float s) const
{
    matrix4 mat;mat*=s;
    return mat;
}

vec4 matrix4::operator*(const vec4& v) const
{
    float x0=v.x();
    float y0=v.y();
    float z0=v.z();
    float w0=v.w();

    return vec4(m[0]*x0+m[4]*y0+m[8]*z0+m[12]*w0,
              m[1]*x0+m[5]*y0+m[9]*z0+m[13]*w0,
              m[2]*x0+m[6]*y0+m[10]*z0+m[14]*w0,
              m[3]*x0+m[7]*y0+m[11]*z0+m[15]*w0);
}
vec3 matrix4::operator*(const vec3& v) const
{
    float x0=v.x();
    float y0=v.y();
    float z0=v.z();

    float w=m[3]*x0+m[7]*y0+m[11]*z0+m[15];

    float epsilon=1e-8;
    if(std::abs(w)<epsilon)
        throw cpe::exception_cpe("Norme=0",MACRO_EXCEPTION_PARAMETER);

    return vec3(m[0]*x0+m[4]*y0+m[8]*z0+m[12],
              m[1]*x0+m[5]*y0+m[9]*z0+m[13],
              m[2]*x0+m[6]*y0+m[10]*z0+m[14])/w;
}

matrix4 matrix4::operator*(const matrix4& m2) const
{
    const float* pm2=m2.m;

    matrix4 current=matrix4::zeros();
    float *p=current.pointer_unprotected();
    for(unsigned int k0=0;k0<4;++k0)
        for(unsigned int k1=0;k1<4;++k1)
            for(unsigned int ks=0;ks<4;++ks)
                p[k0+4*k1] += m[4*ks+k0]*pm2[4*k1+ks];
    return current;
}

matrix4 matrix4::operator/(float s) const
{
    return matrix4(m[0]/s,m[4]/s,m[8]/s,m[12]/s,
                   m[1]/s,m[5]/s,m[9]/s,m[13]/s,
                   m[2]/s,m[6]/s,m[10]/s,m[14]/s,
                   m[3]/s,m[7]/s,m[11]/s,m[15]/s);
}

matrix4& matrix4::operator+=(const matrix4& _m)
{
    const float* pm=_m.m;
    for(unsigned int k=0;k<16;++k)
        m[k] += pm[k];
    return *this;
}

matrix4& matrix4::operator+=(float s)
{
    for(unsigned int k=0;k<16;++k)
        m[k] += s;
    return *this;
}

matrix4& matrix4::operator-=(const matrix4& _m)
{
    const float* pm=_m.m;
    for(unsigned int k=0;k<16;++k)
        m[k] -= pm[k];
    return *this;
}

matrix4& matrix4::operator-=(float s)
{
    for(unsigned int k=0;k<16;++k)
        m[k] -= s;
    return *this;
}

matrix4& matrix4::operator*=(float s)
{
    for(unsigned int k=0;k<16;++k)
        m[k] *= s;
    return *this;
}

matrix4& matrix4::operator/=(float s)
{
    for(unsigned int k=0;k<16;++k)
        m[k] /= s;
    return *this;
}

matrix4 matrix4::operator-() const
{
    return matrix4(-m[0],-m[4],-m[8],-m[12],
                   -m[1],-m[5],-m[9],-m[13],
                   -m[2],-m[6],-m[10],-m[14],
                   -m[3],-m[7],-m[11],-m[15]);
}
matrix4 matrix4::product_compontentwise(const matrix4& _m) const
{
    const float* pm=_m.m;
    return matrix4(m[0]*pm[0],m[4]*pm[4],m[8]*pm[8],m[12]*pm[12],
                   m[1]*pm[1],m[5]*pm[5],m[9]*pm[9],m[13]*pm[13],
                   m[2]*pm[2],m[6]*pm[6],m[10]*pm[10],m[14]*pm[14],
                   m[3]*pm[3],m[7]*pm[7],m[11]*pm[11],m[15]*pm[15]);
}

matrix4& matrix4::product_compontentwise_internal(const matrix4& _m)
{
    const float* pm=_m.m;
    for(unsigned int k=0;k<16;++k)
        m[k]=pm[k];
    return *this;
}

std::ostream& operator<<(std::ostream& stream,const matrix4& _m)
{
    const float* pm=_m.pointer();
    stream<<pm[0]<<","<<pm[4]<<","<<pm[8]<<","<<pm[12]<<"\n"<<
            pm[1]<<","<<pm[5]<<","<<pm[9]<<","<<pm[13]<<"\n"<<
            pm[2]<<","<<pm[6]<<","<<pm[10]<<","<<pm[14]<<"\n"<<
            pm[3]<<","<<pm[7]<<","<<pm[11]<<","<<pm[15]<<"\n";
    return stream;
}

void matrix4::assert_size(const size_t& k1,const size_t& k2) const
{
    if(k1>3 || k2>3)
    {
        std::stringstream stream;
        stream<<"matrix3::assert_size("<<k1<<","<<k2<<") > (3,3)";
        throw exception_cpe(stream.str(),MACRO_EXCEPTION_PARAMETER);
    }
}

const float* matrix4::pointer() const
{
    return m;
}
float* matrix4::pointer_unprotected()
{
    return m;
}
std::vector<float> matrix4::to_vector() const
{
    std::vector<float> m4d(16);
    for(unsigned int kx=0;kx<4;++kx)
        for(unsigned int ky=0;ky<4;++ky)
            m4d[4*ky+kx]=m[4*ky+kx];
    return m4d;
}

matrix4 matrix4::transposed() const
{
    return matrix4(m[0],m[1],m[2],m[3],
                   m[4],m[5],m[6],m[7],
                   m[8],m[9],m[10],m[11],
                   m[12],m[13],m[14],m[15]);
}

matrix4 matrix4::projection_perspective(float fovy,float aspect,float z_near,float z_far)
{
    const float f=1.0/tan(fovy/2.0);
    const float fp=f/aspect;

    const float L=z_near-z_far;
    const float epsilon=1e-8;
    if(std::abs(L)<epsilon)
        throw cpe::exception_cpe("z_far-z_near too small",MACRO_EXCEPTION_PARAMETER);

    const float C=(z_far+z_near)/L;
    const float D=(2*z_far*z_near)/L;

    return matrix4(fp , 0, 0, 0,
                   0  , f, 0, 0,
                   0  , 0, C, D,
                   0  , 0,-1, 0);
}

matrix4 matrix4::projection_frustum(float left,float right,float bottom,float top,float near,float far)
{
    const float width=right-left;
    const float height=top-bottom;
    const float L=far-near;

    const float epsilon=1e-8;
    if(std::abs(width)<epsilon ||
            std::abs(height)<epsilon ||
            std::abs(L)<epsilon)
        throw cpe::exception_cpe("bad camera projection",MACRO_EXCEPTION_PARAMETER);

    const float fw=2*near/width;
    const float fh=2*near/height;
    const float A=(right+left)/width;
    const float B=(top+bottom)/height;
    const float C=-(far+near)/L;
    const float D=-2*far*near/L;

    return matrix4(fw , 0, A, 0,
                   0  ,fh, B, 0,
                   0  , 0, C, D,
                   0  , 0,-1, 0);
}

matrix4 matrix4::projection_orthographic(float left,float right,float bottom,float top,float near,float far)
{
    const float width=right-left;
    const float height=top-bottom;
    const float L=far-near;

    const float epsilon=1e-8;
    if(std::abs(width)<epsilon ||
            std::abs(height)<epsilon ||
            std::abs(L)<epsilon)
        throw cpe::exception_cpe("bad camera projection",MACRO_EXCEPTION_PARAMETER);

    const float sx= 2.0/width;
    const float sy= 2.0/height;
    const float sz=-2.0/L;
    const float tx=-(right+left)/width;
    const float ty=-(top+bottom)/height;
    const float tz=-(far+near)/L;


    return matrix4(sx , 0, 0,tx,
                   0  ,sy, 0,ty,
                   0  , 0,sz,tz,
                   0  , 0, 0, 1);
}

matrix4 matrix4::projection_look_at(const matrix4& current_matrix,const vec3& eye,const vec3& center,const vec3& up)
{
    vec3 f=(center-eye).normalized();
    vec3 s=f.cross(up);
    vec3 u=s.cross(f);

    matrix4 M( s.x(), s.y(), s.z(), 0,
               u.x(), u.y(), u.z(), 0,
              -f.x(),-f.y(),-f.z(), 0,
                  0 ,    0 ,    0 , 1);


    matrix4 T=matrix4::translation(-eye);
    M=M*T;
    M=current_matrix*M;



    return M;
}


matrix4 matrix4::translated(const vec3& translation_param) const
{
    matrix4 temp=*this;
    temp.translate_internal(translation_param);
    return temp;
}

void matrix4::translate_internal(const vec3& translation_param)
{
    m[12]+=translation_param.x();
    m[13]+=translation_param.y();
    m[14]+=translation_param.z();
}

}
